@@ -64,6 +64,9 @@ class IntroduceBufferAuxiliaryVars(CythonTransform):
             # for now...note that pos is wrong 
             raise CompileError(node.pos, "Buffer vars not allowed in module scope")
         for entry in bufvars:
+            if entry.type.dtype.is_ptr:
+                raise CompileError(node.pos, "Buffers with pointer types not yet supported.")
+            
             name = entry.name
             buftype = entry.type
             if buftype.ndim > self.max_ndim:
@@ -291,29 +294,27 @@ def put_assign_to_buffer(lhs_cname, rhs_cname, buffer_aux, buffer_type,
         # Release any existing buffer
         code.putln('__Pyx_SafeReleaseBuffer(&%s);' % bufstruct)
         # Acquire
-        retcode_cname = code.funcstate.allocate_temp(PyrexTypes.c_int_type)
+        retcode_cname = code.funcstate.allocate_temp(PyrexTypes.c_int_type, manage_ref=False)
         code.putln("%s = %s;" % (retcode_cname, getbuffer % rhs_cname))
-        code.putln('if (%s) ' % (code.unlikely("%s < 0" % retcode_cname)))
+        code.putln('if (%s) {' % (code.unlikely("%s < 0" % retcode_cname)))
         # If acquisition failed, attempt to reacquire the old buffer
         # before raising the exception. A failure of reacquisition
         # will cause the reacquisition exception to be reported, one
         # can consider working around this later.
-        code.begin_block()
-        type, value, tb = [code.funcstate.allocate_temp(PyrexTypes.py_object_type)
+        type, value, tb = [code.funcstate.allocate_temp(PyrexTypes.py_object_type, manage_ref=False)
                            for i in range(3)]
         code.putln('PyErr_Fetch(&%s, &%s, &%s);' % (type, value, tb))
-        code.put('if (%s) ' % code.unlikely("%s == -1" % (getbuffer % lhs_cname)))
-        code.begin_block()
-        code.putln('Py_XDECREF(%s); Py_XDECREF(%s); Py_XDECREF(%s);' % (type, value, tb))
+        code.putln('if (%s) {' % code.unlikely("%s == -1" % (getbuffer % lhs_cname)))
+        code.putln('Py_XDECREF(%s); Py_XDECREF(%s); Py_XDECREF(%s);' % (type, value, tb)) # Do not refnanny these!
         code.globalstate.use_utility_code(raise_buffer_fallback_code)
         code.putln('__Pyx_RaiseBufferFallbackError();')
         code.putln('} else {')
         code.putln('PyErr_Restore(%s, %s, %s);' % (type, value, tb))
         for t in (type, value, tb):
             code.funcstate.release_temp(t)
-        code.end_block()
+        code.putln('}')
+        code.putln('}')
         # Unpack indices
-        code.end_block()
         put_unpack_buffer_aux_into_scope(buffer_aux, buffer_type.mode, code)
         code.putln(code.error_goto_if_neg(retcode_cname, pos))
         code.funcstate.release_temp(retcode_cname)
@@ -322,7 +323,7 @@ def put_assign_to_buffer(lhs_cname, rhs_cname, buffer_aux, buffer_type,
         # In this case, auxiliary vars should be set up right in initialization to a zero-buffer,
         # so it suffices to set the buf field to NULL.
         code.putln('if (%s) {' % code.unlikely("%s == -1" % (getbuffer % rhs_cname)))
-        code.putln('%s = %s; Py_INCREF(Py_None); %s.buf = NULL;' %
+        code.putln('%s = %s; __Pyx_INCREF(Py_None); %s.buf = NULL;' %
                    (lhs_cname,
                     PyrexTypes.typecast(buffer_type, PyrexTypes.py_object_type, "Py_None"),
                     bufstruct))
@@ -355,7 +356,7 @@ def put_buffer_lookup_code(entry, index_signeds, index_cnames, options, pos, cod
         # We allocate a temporary which is initialized to -1, meaning OK (!).
         # If an error occurs, the temp is set to the dimension index the
         # error is occuring at.
-        tmp_cname = code.funcstate.allocate_temp(PyrexTypes.c_int_type)
+        tmp_cname = code.funcstate.allocate_temp(PyrexTypes.c_int_type, manage_ref=False)
         code.putln("%s = -1;" % tmp_cname)
         for dim, (signed, cname, shape) in enumerate(zip(index_signeds, index_cnames,
                                                          bufaux.shapevars)):
@@ -374,11 +375,10 @@ def put_buffer_lookup_code(entry, index_signeds, index_cnames, options, pos, cod
                 code.unlikely("%s >= %s" % (cname, shape.cname)),
                 tmp_cname, dim))
         code.globalstate.use_utility_code(raise_indexerror_code)
-        code.put("if (%s) " % code.unlikely("%s != -1" % tmp_cname))
-        code.begin_block()
+        code.putln("if (%s) {" % code.unlikely("%s != -1" % tmp_cname))
         code.putln('__Pyx_RaiseBufferIndexError(%s);' % tmp_cname)
         code.putln(code.error_goto(pos))
-        code.end_block()
+        code.putln('}')
         code.funcstate.release_temp(tmp_cname)
     elif negative_indices:
         # Only fix negative indices.
@@ -500,7 +500,7 @@ def mangle_dtype_name(dtype):
     elif dtype.is_ptr:
         return "ptr"
     else:
-        if dtype.typestring is None:
+        if dtype.is_typedef or dtype.is_struct_or_union:
             prefix = "nn_"
         else:
             prefix = ""
@@ -552,10 +552,8 @@ def create_typestringchecker(protocode, defcode, name, dtype):
         defcode.putln("int ok;")
         defcode.putln("ts = __Pyx_ConsumeWhitespace(ts); if (!ts) return NULL;")
         defcode.putln("if (*ts == '1') ++ts;")
-        if dtype.typestring is not None:
-            assert len(dtype.typestring) == 1
-            # Can use direct comparison
-            defcode.putln("ok = (*ts == '%s');" % dtype.typestring)
+        if dtype.is_pyobject:
+            defcode.putln("ok = (*ts == 'O');")
         else:
             # Cannot trust declared size; but rely on int vs float and
             # signed/unsigned to be correctly declared. Use a switch statement
