@@ -10,10 +10,12 @@ from PyrexTypes import py_object_type, typecast
 from TypeSlots import method_coexist
 from Scanning import SourceDescriptor
 from Cython.StringIOTree import StringIOTree
+import DebugFlags
 try:
     set
 except NameError:
     from sets import Set as set
+import DebugFlags
 
 class FunctionState(object):
     # return_label     string          function return point label
@@ -26,8 +28,9 @@ class FunctionState(object):
     # exc_vars         (string * 3)    exception variables for reraise, or None
 
     # Not used for now, perhaps later
-    def __init__(self, names_taken=set()):
+    def __init__(self, owner, names_taken=set()):
         self.names_taken = names_taken
+        self.owner = owner
         
         self.error_label = None
         self.label_counter = 0
@@ -40,9 +43,9 @@ class FunctionState(object):
         self.in_try_finally = 0
         self.exc_vars = None
 
-        self.temps_allocated = [] # of (name, type)
-        self.temps_free = {} # type -> list of free vars
-        self.temps_used_type = {} # name -> type
+        self.temps_allocated = [] # of (name, type, manage_ref)
+        self.temps_free = {} # (type, manage_ref) -> list of free vars with same type/managed status
+        self.temps_used_type = {} # name -> (type, manage_ref)
         self.temp_counter = 0
 
     def new_label(self, name=None):
@@ -70,8 +73,8 @@ class FunctionState(object):
     def new_loop_labels(self):
         old_labels = self.get_loop_labels()
         self.set_loop_labels(
-            (self.new_label(), 
-             self.new_label()))
+            (self.new_label("continue"), 
+             self.new_label("break")))
         return old_labels
     
     def get_all_labels(self):
@@ -104,15 +107,28 @@ class FunctionState(object):
     def label_used(self, lbl):
         return lbl in self.labels_used
 
-    def allocate_temp(self, type):
+    def allocate_temp(self, type, manage_ref):
         """
         Allocates a temporary (which may create a new one or get a previously
         allocated and released one of the same type). Type is simply registered
         and handed back, but will usually be a PyrexType.
 
+        If type.is_pyobject, manage_ref comes into play. If manage_ref is set to
+        True, the temp will be decref-ed on return statements and in exception
+        handling clauses. Otherwise the caller has to deal with any reference
+        counting of the variable.
+
+        If not type.is_pyobject, then manage_ref will be ignored, but it
+        still has to be passed. It is recommended to pass False by convention
+        if it is known that type will never be a Python object.
+
         A C string referring to the variable is returned.
         """
-        freelist = self.temps_free.get(type)
+        if not type.is_pyobject:
+            # Make manage_ref canonical, so that manage_ref will always mean
+            # a decref is needed.
+            manage_ref = False
+        freelist = self.temps_free.get((type, manage_ref))
         if freelist is not None and len(freelist) > 0:
             result = freelist.pop()
         else:
@@ -120,8 +136,10 @@ class FunctionState(object):
                 self.temp_counter += 1
                 result = "%s%d" % (Naming.codewriter_temp_prefix, self.temp_counter)
                 if not result in self.names_taken: break
-            self.temps_allocated.append((result, type))
-        self.temps_used_type[result] = type
+            self.temps_allocated.append((result, type, manage_ref))
+        self.temps_used_type[result] = (type, manage_ref)
+        if DebugFlags.debug_temp_code_comments:
+            self.owner.putln("/* %s allocated */" % result)
         return result
 
     def release_temp(self, name):
@@ -129,13 +147,54 @@ class FunctionState(object):
         Releases a temporary so that it can be reused by other code needing
         a temp of the same type.
         """
-        type = self.temps_used_type[name]
-        freelist = self.temps_free.get(type)
+        type, manage_ref = self.temps_used_type[name]
+        freelist = self.temps_free.get((type, manage_ref))
         if freelist is None:
             freelist = []
-
-            self.temps_free[type] = freelist
+            self.temps_free[(type, manage_ref)] = freelist
+        if name in freelist:
+            raise RuntimeError("Temp %s freed twice!" % name)
         freelist.append(name)
+        if DebugFlags.debug_temp_code_comments:
+            self.owner.putln("/* %s released */" % name)
+
+    def temps_in_use(self):
+        """Return a list of (cname,type,manage_ref) tuples of temp names and their type
+        that are currently in use.
+        """
+        used = []
+        for name, type, manage_ref in self.temps_allocated:
+            freelist = self.temps_free.get((type, manage_ref))
+            if freelist is None or name not in freelist:
+                used.append((name, type, manage_ref))
+        return used
+
+    def temps_holding_reference(self):
+        """Return a list of (cname,type) tuples of temp names and their type
+        that are currently in use. This includes only temps of a
+        Python object type which owns its reference.
+        """
+        return [(name, type)
+                for name, type, manage_ref in self.temps_in_use()
+                if manage_ref]
+
+    def all_managed_temps(self):
+        """Return a list of (cname, type) tuples of refcount-managed Python objects.
+        """
+        return [(cname, type)
+                for cname, type, manage_ref in self.temps_allocated
+                if manage_ref]
+
+    def all_free_managed_temps(self):
+        """Return a list of (cname, type) tuples of refcount-managed Python
+        objects that are not currently in use.  This is used by
+        try-except and try-finally blocks to clean up temps in the
+        error case.
+        """
+        return [(cname, type)
+                for (type, manage_ref), freelist in self.temps_free.iteritems()
+                if manage_ref
+                for cname in freelist]
 
 class GlobalState(object):
     # filename_table   {string : int}  for finding filename table indexes
@@ -161,13 +220,21 @@ class GlobalState(object):
     # interned_nums
     # cached_builtins
 
-    def __init__(self, rootwriter):
+    # directives       set             Temporary variable used to track
+    #                                  the current set of directives in the code generation
+    #                                  process.
+
+    directives = {}
+
+    def __init__(self, rootwriter, emit_linenums=False):
         self.filename_table = {}
         self.filename_list = []
         self.input_file_contents = {}
         self.used_utility_code = set()
         self.declared_cnames = {}
         self.pystring_table_needed = False
+        self.in_utility_code_generation = False
+        self.emit_linenums = emit_linenums
 
     def initwriters(self, rootwriter):
         self.utilprotowriter = rootwriter.new_writer()
@@ -176,19 +243,23 @@ class GlobalState(object):
         self.pystring_table = rootwriter.new_writer()
         self.init_cached_builtins_writer = rootwriter.new_writer()
         self.initwriter = rootwriter.new_writer()
+        self.cleanupwriter = rootwriter.new_writer()
 
         if Options.cache_builtins:
             self.init_cached_builtins_writer.enter_cfunc_scope()
             self.init_cached_builtins_writer.putln("static int __Pyx_InitCachedBuiltins(void) {")
 
         self.initwriter.enter_cfunc_scope()
-        self.initwriter.putln("").putln("static int __Pyx_InitGlobals(void) {")
+        self.initwriter.putln("")
+        self.initwriter.putln("static int __Pyx_InitGlobals(void) {")
 
-        (self.pystring_table
-         .putln("")
-         .putln("static __Pyx_StringTabEntry %s[] = {" %
+        self.cleanupwriter.enter_cfunc_scope()
+        self.cleanupwriter.putln("")
+        self.cleanupwriter.putln("static void __Pyx_CleanupGlobals(void) {")
+
+        self.pystring_table.putln("")
+        self.pystring_table.putln("static __Pyx_StringTabEntry %s[] = {" %
                 Naming.stringtab_cname)
-        )
 
     #
     # Global constants, interned objects, etc.
@@ -200,7 +271,8 @@ class GlobalState(object):
         # This is called when it is known that no more global declarations will
         # declared (but can be called before or after insert_XXX).
         if self.pystring_table_needed:
-            self.pystring_table.putln("{0, 0, 0, 0, 0, 0}").putln("};")
+            self.pystring_table.putln("{0, 0, 0, 0, 0, 0}")
+            self.pystring_table.putln("};")
             import Nodes
             self.use_utility_code(Nodes.init_string_tab_utility_code)
             self.initwriter.putln(
@@ -209,21 +281,23 @@ class GlobalState(object):
                     self.initwriter.error_goto(self.module_pos)))
 
         if Options.cache_builtins:
-            (self.init_cached_builtins_writer
-             .putln("return 0;")
-             .put_label(self.init_cached_builtins_writer.error_label)
-             .putln("return -1;")
-             .putln("}")
-             .exit_cfunc_scope()
-             )
+            w = self.init_cached_builtins_writer
+            w.putln("return 0;")
+            w.put_label(w.error_label)
+            w.putln("return -1;")
+            w.putln("}")
+            w.exit_cfunc_scope()
 
-        (self.initwriter
-         .putln("return 0;")
-         .put_label(self.initwriter.error_label)
-         .putln("return -1;")
-         .putln("}")
-         .exit_cfunc_scope()
-         )
+        w = self.initwriter
+        w.putln("return 0;")
+        w.put_label(w.error_label)
+        w.putln("return -1;")
+        w.putln("}")
+        w.exit_cfunc_scope()
+
+        w = self.cleanupwriter
+        w.putln("}")
+        w.exit_cfunc_scope()
          
     def insert_initcode_into(self, code):
         if self.pystring_table_needed:
@@ -232,6 +306,9 @@ class GlobalState(object):
             code.insert(self.init_cached_builtins_writer)
         code.insert(self.initwriter)
 
+    def insert_cleanupcode_into(self, code):
+        code.insert(self.cleanupwriter)
+
     def put_pyobject_decl(self, entry):
         self.decls_writer.putln("static PyObject *%s;" % entry.cname)
 
@@ -243,7 +320,7 @@ class GlobalState(object):
     def should_declare(self, cname, entry):
         if cname in self.declared_cnames:
             other = self.declared_cnames[cname]
-            assert entry.type == other.type
+            assert str(entry.type) == str(other.type)
             assert entry.init == other.init
             return False
         else:
@@ -275,9 +352,9 @@ class GlobalState(object):
     def add_interned_num_decl(self, entry):
         if self.should_declare(entry.cname, entry):
             if entry.init[-1] == "L":
-                self.initwriter.putln('%s = PyLong_FromString("%s", 0, 0); %s;' % (
+                self.initwriter.putln('%s = PyLong_FromString((char *)"%s", 0, 0); %s;' % (
                     entry.cname,
-                    entry.init,
+                    entry.init[:-1], # strip 'L' for Py3 compatibility
                     self.initwriter.error_goto_if_null(entry.cname, self.module_pos)))
             else:
                 self.initwriter.putln("%s = PyInt_FromLong(%s); %s;" % (
@@ -318,6 +395,8 @@ class GlobalState(object):
         except KeyError:
             F = [u' * ' + line.rstrip().replace(
                     u'*/', u'*[inserted by cython to avoid comment closer]/'
+                    ).replace(
+                    u'/*', u'/[inserted by cython to avoid comment start]*'
                     ).encode('ASCII', 'replace') # + Py2 auto-decode to unicode
                  for line in source_desc.get_lines()]
             if len(F) == 0: F.append(u'')
@@ -328,7 +407,7 @@ class GlobalState(object):
     # Utility code state
     #
     
-    def use_utility_code(self, codetup, name=None):
+    def use_utility_code(self, utility_code, name=None):
         """
         Adds the given utility code to the C file if needed.
 
@@ -338,16 +417,19 @@ class GlobalState(object):
         If name is provided, it is used as an identifier to avoid inserting
         code twice. Otherwise, id(codetup) is used as such an identifier.
         """
-        if name is None: name = id(codetup)
+        if name is None: name = id(utility_code)
         if self.check_utility_code_needed_and_register(name):
-            proto, _def = codetup
-            self.utilprotowriter.put(proto)
-            self.utildefwriter.put(_def)
+            if utility_code.proto:
+                self.utilprotowriter.put(utility_code.proto)
+            if utility_code.impl:
+                self.utildefwriter.put(utility_code.impl)
+            utility_code.write_init_code(self.initwriter, self.module_pos)
+            utility_code.write_cleanup_code(self.cleanupwriter, self.module_pos)
 
-    def has_utility_code(self, name):
+    def has_code(self, name):
         return name in self.used_utility_code
 
-    def use_generated_code(self, func, name, *args, **kw):
+    def use_code_from(self, func, name, *args, **kw):
         """
         Requests that the utility code that func can generate is used in the C
         file. func is called like this:
@@ -378,6 +460,8 @@ class GlobalState(object):
         writer.insert(self.utilprotowriter)
 
     def put_utility_code_defs(self, writer):
+        if self.emit_linenums:
+            writer.write('\n#line 1 "cython_utility"\n')
         writer.insert(self.utildefwriter)
 
 
@@ -403,7 +487,7 @@ class CCodeWriter(object):
     - marker: Not copied to insertion point
     - filename_table, filename_list, input_file_contents: All codewriters
       coming from the same root share the same instances simultaneously.
-    """ 
+    """
     
     # f                file            output file
     # buffer           StringIOTree
@@ -415,19 +499,22 @@ class CCodeWriter(object):
     #                                  generation (labels and temps state etc.)
     # globalstate      GlobalState     contains state global for a C file (input file info,
     #                                  utility code, declared constants etc.)
-   
-    def __init__(self, create_from=None, buffer=None, copy_formatting=False):
+    # emit_linenums    boolean         whether or not to write #line pragmas 
+    
+    def __init__(self, create_from=None, buffer=None, copy_formatting=False, emit_linenums=None):
         if buffer is None: buffer = StringIOTree()
         self.buffer = buffer
         self.marker = None
         self.last_marker_line = 0
+        self.source_desc = ""
         
         self.funcstate = None
         self.level = 0
+        self.call_level = 0
         self.bol = 1
         if create_from is None:
             # Root CCodeWriter
-            self.globalstate = GlobalState(self)
+            self.globalstate = GlobalState(self, emit_linenums=emit_linenums)
             self.globalstate.initwriters(self)
             # ^^^ need seperate step because this will reference self.globalstate
         else:
@@ -437,6 +524,11 @@ class CCodeWriter(object):
             if copy_formatting:
                 self.level = create_from.level
                 self.bol = create_from.bol
+                self.call_level = create_from.call_level
+        if emit_linenums is None:
+            self.emit_linenums = self.globalstate.emit_linenums
+        else:
+            self.emit_linenums = emit_linenums
 
     def create_new(self, create_from, buffer, copy_formatting):
         # polymorphic constructor -- very slightly more versatile
@@ -496,7 +588,7 @@ class CCodeWriter(object):
 
 
     def enter_cfunc_scope(self):
-        self.funcstate = FunctionState()
+        self.funcstate = FunctionState(self)
     
     def exit_cfunc_scope(self):
         self.funcstate = None
@@ -504,11 +596,12 @@ class CCodeWriter(object):
     def putln(self, code = ""):
         if self.marker and self.bol:
             self.emit_marker()
+        if self.emit_linenums and self.last_marker_line != 0:
+            self.write('\n#line %s "%s"\n' % (self.last_marker_line, self.source_desc))
         if code:
             self.put(code)
         self.write("\n");
         self.bol = 1
-        return self
     
     def emit_marker(self):
         self.write("\n");
@@ -516,7 +609,6 @@ class CCodeWriter(object):
         self.write("/* %s */\n" % self.marker[1])
         self.last_marker_line = self.marker[0]
         self.marker = None
-        return self
 
     def put_safe(self, code):
         # put code, but ignore {}
@@ -525,12 +617,18 @@ class CCodeWriter(object):
 
     def put(self, code):
         fix_indent = False
-        dl = code.count("{") - code.count("}")
-        if dl < 0:
-            self.level += dl
-        elif dl == 0 and code.startswith('}'):
-            fix_indent = True
-            self.level -= 1
+        if "{" in code:
+            dl = code.count("{")
+        else:
+            dl = 0
+        if "}" in code:
+            dl -= code.count("}")
+            if dl < 0:
+                self.level += dl
+            elif dl == 0 and code[0] == "}":
+                # special cases like "} else {" need a temporary dedent
+                fix_indent = True
+                self.level -= 1
         if self.bol:
             self.indent()
         self.write(code)
@@ -539,25 +637,20 @@ class CCodeWriter(object):
             self.level += dl
         elif fix_indent:
             self.level += 1
-        return self
 
     def increase_indent(self):
         self.level = self.level + 1
-        return self
     
     def decrease_indent(self):
         self.level = self.level - 1
-        return self
     
     def begin_block(self):
         self.putln("{")
         self.increase_indent()
-        return self
     
     def end_block(self):
         self.decrease_indent()
         self.putln("}")
-        return self
     
     def indent(self):
         self.write("  " * self.level)
@@ -580,17 +673,16 @@ class CCodeWriter(object):
         marker = u'"%s":%d\n%s\n' % (
             source_desc.get_escaped_description(), line, u'\n'.join(lines))
         self.marker = (line, marker)
-
+        if self.emit_linenums:
+            self.source_desc = source_desc.get_escaped_description()
         
     def put_label(self, lbl):
         if lbl in self.funcstate.labels_used:
             self.putln("%s:;" % lbl)
-        return self
     
     def put_goto(self, lbl):
         self.funcstate.use_label(lbl)
         self.putln("goto %s;" % lbl)
-        return self
     
     def put_var_declarations(self, entries, static = 0, dll_linkage = None,
             definition = True):
@@ -630,7 +722,7 @@ class CCodeWriter(object):
         self.putln(";")
 
     def put_temp_declarations(self, func_context):
-        for name, type in func_context.temps_allocated:
+        for name, type, manage_ref in func_context.temps_allocated:
             decl = type.declaration_code(name)
             if type.is_pyobject:
                 self.putln("%s = NULL;" % decl)
@@ -639,8 +731,8 @@ class CCodeWriter(object):
 
     def entry_as_pyobject(self, entry):
         type = entry.type
-        if (not entry.is_self_arg and not entry.type.is_complete()) \
-            or (entry.type.is_extension_type and entry.type.base_type):
+        if (not entry.is_self_arg and not entry.type.is_complete()
+            or entry.type.is_extension_type):
             return "(PyObject *)" + entry.cname
         else:
             return entry.cname
@@ -648,47 +740,91 @@ class CCodeWriter(object):
     def as_pyobject(self, cname, type):
         return typecast(py_object_type, type, cname)
     
-    def put_incref(self, cname, type):
-        self.putln("Py_INCREF(%s);" % self.as_pyobject(cname, type))
+    def put_gotref(self, cname):
+        self.putln("__Pyx_GOTREF(%s);" % cname)
     
-    def put_decref(self, cname, type):
-        self.putln("Py_DECREF(%s);" % self.as_pyobject(cname, type))
+    def put_giveref(self, cname):
+        self.putln("__Pyx_GIVEREF(%s);" % cname)
     
+    def put_xgiveref(self, cname):
+        self.putln("__Pyx_XGIVEREF(%s);" % cname)
+
+    def put_xgotref(self, cname):
+        self.putln("__Pyx_XGOTREF(%s);" % cname)
+
+    def put_incref(self, cname, type, nanny=True):
+        if nanny:
+            self.putln("__Pyx_INCREF(%s);" % self.as_pyobject(cname, type))
+        else:
+            self.putln("Py_INCREF(%s);" % self.as_pyobject(cname, type))
+    
+    def put_decref(self, cname, type, nanny=True):
+        if nanny:
+            self.putln("__Pyx_DECREF(%s);" % self.as_pyobject(cname, type))
+        else:
+            self.putln("Py_DECREF(%s);" % self.as_pyobject(cname, type))
+
+    def put_var_gotref(self, entry):
+        if entry.type.is_pyobject:
+            self.putln("__Pyx_GOTREF(%s);" % self.entry_as_pyobject(entry))
+        
+    def put_var_giveref(self, entry):
+        if entry.type.is_pyobject:
+            self.putln("__Pyx_GIVEREF(%s);" % self.entry_as_pyobject(entry))
+
+    def put_var_xgotref(self, entry):
+        if entry.type.is_pyobject:
+            self.putln("__Pyx_XGOTREF(%s);" % self.entry_as_pyobject(entry))
+
+    def put_var_xgiveref(self, entry):
+        if entry.type.is_pyobject:
+            self.putln("__Pyx_XGIVEREF(%s);" % self.entry_as_pyobject(entry))
+
     def put_var_incref(self, entry):
         if entry.type.is_pyobject:
-            self.putln("Py_INCREF(%s);" % self.entry_as_pyobject(entry))
+            self.putln("__Pyx_INCREF(%s);" % self.entry_as_pyobject(entry))
     
-    def put_decref_clear(self, cname, type):
-        self.putln("Py_DECREF(%s); %s = 0;" % (
-            typecast(py_object_type, type, cname), cname))
-            #self.as_pyobject(cname, type), cname))
+    def put_decref_clear(self, cname, type, nanny=True):
+        if nanny:
+            self.putln("__Pyx_DECREF(%s); %s = 0;" % (
+                typecast(py_object_type, type, cname), cname))
+        else:
+            self.putln("Py_DECREF(%s); %s = 0;" % (
+                typecast(py_object_type, type, cname), cname))
     
-    def put_xdecref(self, cname, type):
-        self.putln("Py_XDECREF(%s);" % self.as_pyobject(cname, type))
+    def put_xdecref(self, cname, type, nanny=True):
+        if nanny:
+            self.putln("__Pyx_XDECREF(%s);" % self.as_pyobject(cname, type))
+        else:
+            self.putln("Py_XDECREF(%s);" % self.as_pyobject(cname, type))
     
-    def put_xdecref_clear(self, cname, type):
-        self.putln("Py_XDECREF(%s); %s = 0;" % (
-            self.as_pyobject(cname, type), cname))
+    def put_xdecref_clear(self, cname, type, nanny=True):
+        if nanny:
+            self.putln("__Pyx_XDECREF(%s); %s = 0;" % (
+                self.as_pyobject(cname, type), cname))
+        else:
+            self.putln("Py_XDECREF(%s); %s = 0;" % (
+                self.as_pyobject(cname, type), cname))
 
     def put_var_decref(self, entry):
         if entry.type.is_pyobject:
             if entry.init_to_none is False:
-                self.putln("Py_XDECREF(%s);" % self.entry_as_pyobject(entry))
+                self.putln("__Pyx_XDECREF(%s);" % self.entry_as_pyobject(entry))
             else:
-                self.putln("Py_DECREF(%s);" % self.entry_as_pyobject(entry))
+                self.putln("__Pyx_DECREF(%s);" % self.entry_as_pyobject(entry))
     
     def put_var_decref_clear(self, entry):
         if entry.type.is_pyobject:
-            self.putln("Py_DECREF(%s); %s = 0;" % (
+            self.putln("__Pyx_DECREF(%s); %s = 0;" % (
                 self.entry_as_pyobject(entry), entry.cname))
     
     def put_var_xdecref(self, entry):
         if entry.type.is_pyobject:
-            self.putln("Py_XDECREF(%s);" % self.entry_as_pyobject(entry))
+            self.putln("__Pyx_XDECREF(%s);" % self.entry_as_pyobject(entry))
     
     def put_var_xdecref_clear(self, entry):
         if entry.type.is_pyobject:
-            self.putln("Py_XDECREF(%s); %s = 0;" % (
+            self.putln("__Pyx_XDECREF(%s); %s = 0;" % (
                 self.entry_as_pyobject(entry), entry.cname))
     
     def put_var_decrefs(self, entries, used_only = 0):
@@ -707,15 +843,18 @@ class CCodeWriter(object):
         for entry in entries:
             self.put_var_xdecref_clear(entry)
     
-    def put_init_to_py_none(self, cname, type):
+    def put_init_to_py_none(self, cname, type, nanny=True):
         py_none = typecast(type, py_object_type, "Py_None")
-        self.putln("%s = %s; Py_INCREF(Py_None);" % (cname, py_none))
+        if nanny:
+            self.putln("%s = %s; __Pyx_INCREF(Py_None);" % (cname, py_none))
+        else:
+            self.putln("%s = %s; Py_INCREF(Py_None);" % (cname, py_none))
     
-    def put_init_var_to_py_none(self, entry, template = "%s"):
+    def put_init_var_to_py_none(self, entry, template = "%s", nanny=True):
         code = template % entry.cname
         #if entry.type.is_extension_type:
-        #	code = "((PyObject*)%s)" % code
-        self.put_init_to_py_none(code, entry.type)
+        #    code = "((PyObject*)%s)" % code
+        self.put_init_to_py_none(code, entry.type, nanny)
 
     def put_pymethoddef(self, entry, term):
         if entry.doc:
@@ -727,13 +866,13 @@ class CCodeWriter(object):
             if entry.is_special:
                 method_flags += [method_coexist]
             self.putln(
-                '{"%s", (PyCFunction)%s, %s, %s}%s' % (
+                '{__Pyx_NAMESTR("%s"), (PyCFunction)%s, %s, __Pyx_DOCSTR(%s)}%s' % (
                     entry.name, 
                     entry.func_cname,
                     "|".join(method_flags),
                     doc_code,
                     term))
-    
+
     def put_error_if_neg(self, pos, value):
 #        return self.putln("if (unlikely(%s < 0)) %s" % (value, self.error_goto(pos)))  # TODO this path is almost _never_ taken, yet this macro makes is slower!
         return self.putln("if (%s < 0) %s" % (value, self.error_goto(pos)))
@@ -747,21 +886,25 @@ class CCodeWriter(object):
             return 'unlikely(%s)' % cond
         else:
             return cond
-        
-    def error_goto(self, pos):
-        lbl = self.funcstate.error_label
-        self.funcstate.use_label(lbl)
+
+    def set_error_info(self, pos):
         if Options.c_line_in_traceback:
             cinfo = " %s = %s;" % (Naming.clineno_cname, Naming.line_c_macro)
         else:
             cinfo = ""
-        return "{%s = %s[%s]; %s = %s;%s goto %s;}" % (
+        return "%s = %s[%s]; %s = %s;%s" % (
             Naming.filename_cname,
             Naming.filetable_cname,
             self.lookup_filename(pos[0]),
             Naming.lineno_cname,
             pos[1],
-            cinfo,
+            cinfo)
+        
+    def error_goto(self, pos):
+        lbl = self.funcstate.error_label
+        self.funcstate.use_label(lbl)
+        return "{%s goto %s;}" % (
+            self.set_error_info(pos),
             lbl)
 
     def error_goto_if(self, cond, pos):
@@ -779,6 +922,12 @@ class CCodeWriter(object):
     def lookup_filename(self, filename):
         return self.globalstate.lookup_filename(filename)
 
+    def put_setup_refcount_context(self, name):
+        self.putln('__Pyx_SetupRefcountContext("%s");' % name)
+
+    def put_finish_refcount_context(self):
+        self.putln("__Pyx_FinishRefcountContext();")
+
 
 class PyrexCodeWriter:
     # f                file      output file
