@@ -80,12 +80,22 @@ class Context:
         from ParseTreeTransforms import AnalyseDeclarationsTransform, AnalyseExpressionsTransform
         from ParseTreeTransforms import CreateClosureClasses, MarkClosureVisitor, DecoratorTransform
         from ParseTreeTransforms import InterpretCompilerDirectives, TransformBuiltinMethods
-        from ParseTreeTransforms import AlignFunctionDefinitions
+        from ParseTreeTransforms import ComprehensionTransform, AlignFunctionDefinitions
         from AutoDocTransforms import EmbedSignature
-        from Optimize import FlattenInListTransform, SwitchTransform, FinalOptimizePhase
+        from Optimize import FlattenInListTransform, SwitchTransform, IterationTransform
+        from Optimize import FlattenBuiltinTypeCreation, ConstantFolding, FinalOptimizePhase
         from Buffer import IntroduceBufferAuxiliaryVars
         from ModuleNode import check_c_declarations
 
+        # Temporary hack that can be used to ensure that all result_code's
+        # are generated at code generation time.
+        import Visitor
+        class ClearResultCodes(Visitor.CythonTransform):
+            def visit_ExprNode(self, node):
+                self.visitchildren(node)
+                node.result_code = "<cleared>"
+                return node
+
         if pxd:
             _check_c_declarations = None
             _specific_post_parse = PxdPostParse(self)
@@ -102,8 +112,8 @@ class Context:
             NormalizeTree(self),
             PostParse(self),
             _specific_post_parse,
-            _align_function_definitions,
             InterpretCompilerDirectives(self, self.pragma_overrides),
+            _align_function_definitions,
             FlattenInListTransform(),
             WithTransform(self),
             DecoratorTransform(self),
@@ -113,8 +123,13 @@ class Context:
             IntroduceBufferAuxiliaryVars(self),
             _check_c_declarations,
             AnalyseExpressionsTransform(self),
+            FlattenBuiltinTypeCreation(),
+            ConstantFolding(),
+#            ComprehensionTransform(),
+            IterationTransform(),
             SwitchTransform(),
             FinalOptimizePhase(self),
+#            ClearResultCodes(self),
 #            SpecialFunctions(self),
             #        CreateClosureClasses(context),
             ]
