@@ -1,9 +1,9 @@
-from Cython.Compiler.Visitor import VisitorTransform, CythonTransform
+from Cython.Compiler.Visitor import VisitorTransform, CythonTransform, TreeVisitor
 from Cython.Compiler.ModuleNode import ModuleNode
 from Cython.Compiler.Nodes import *
 from Cython.Compiler.ExprNodes import *
 from Cython.Compiler.UtilNodes import *
-from Cython.Compiler.TreeFragment import TreeFragment
+from Cython.Compiler.TreeFragment import TreeFragment, TemplateTransform
 from Cython.Compiler.StringEncoding import EncodedString
 from Cython.Compiler.Errors import CompileError
 try:
@@ -12,6 +12,48 @@ except NameError:
     from sets import Set as set
 import copy
 
+
+class NameNodeCollector(TreeVisitor):
+    """Collect all NameNodes of a (sub-)tree in the ``name_nodes``
+    attribute.
+    """
+    def __init__(self):
+        super(NameNodeCollector, self).__init__()
+        self.name_nodes = []
+
+    visit_Node = TreeVisitor.visitchildren
+
+    def visit_NameNode(self, node):
+        self.name_nodes.append(node)
+
+
+class SkipDeclarations:
+    """
+    Variable and function declarations can often have a deep tree structure, 
+    and yet most transformations don't need to descend to this depth. 
+    
+    Declaration nodes are removed after AnalyseDeclarationsTransform, so there 
+    is no need to use this for transformations after that point. 
+    """
+    def visit_CTypeDefNode(self, node):
+        return node
+    
+    def visit_CVarDefNode(self, node):
+        return node
+    
+    def visit_CDeclaratorNode(self, node):
+        return node
+    
+    def visit_CBaseTypeNode(self, node):
+        return node
+    
+    def visit_CEnumDefNode(self, node):
+        return node
+
+    def visit_CStructOrUnionDefNode(self, node):
+        return node
+
+
 class NormalizeTree(CythonTransform):
     """
     This transform fixes up a few things after parsing
@@ -77,6 +119,9 @@ class NormalizeTree(CythonTransform):
         else:
             return []
 
+    def visit_CDeclaratorNode(self, node):
+        return node    
+
 
 class PostParseError(CompileError): pass
 
@@ -202,7 +247,7 @@ class PostParse(CythonTransform):
             raise PostParseError(node.pos, ERR_BUF_LOCALONLY)
         return node
 
-class PxdPostParse(CythonTransform):
+class PxdPostParse(CythonTransform, SkipDeclarations):
     """
     Basic interpretation/validity checking that should only be
     done on pxd trees.
@@ -212,8 +257,12 @@ class PxdPostParse(CythonTransform):
 
     - "def" functions are let through only if they fill the
     getbuffer/releasebuffer slots
+    
+    - cdef functions are let through only if they are on the
+    top level and are declared "inline"
     """
-    ERR_FUNCDEF_NOT_ALLOWED = 'function definition not allowed here'
+    ERR_INLINE_ONLY = "function definition in pxd file must be declared 'cdef inline'"
+    ERR_NOGO_WITH_INLINE = "inline function definition in pxd file cannot be '%s'"
 
     def __call__(self, node):
         self.scope_type = 'pxd'
@@ -229,20 +278,31 @@ class PxdPostParse(CythonTransform):
     def visit_FuncDefNode(self, node):
         # FuncDefNode always come with an implementation (without
         # an imp they are CVarDefNodes..)
-        ok = False
+        err = self.ERR_INLINE_ONLY
 
         if (isinstance(node, DefNode) and self.scope_type == 'cclass'
             and node.name in ('__getbuffer__', '__releasebuffer__')):
-            ok = True
+            err = None # allow these slots
+            
+        if isinstance(node, CFuncDefNode):
+            if u'inline' in node.modifiers and self.scope_type == 'pxd':
+                node.inline_in_pxd = True
+                if node.visibility != 'private':
+                    err = self.ERR_NOGO_WITH_INLINE % node.visibility
+                elif node.api:
+                    err = self.ERR_NOGO_WITH_INLINE % 'api'
+                else:
+                    err = None # allow inline function
+            else:
+                err = self.ERR_INLINE_ONLY
 
-        if not ok:
-            self.context.nonfatal_error(PostParseError(node.pos,
-                self.ERR_FUNCDEF_NOT_ALLOWED))
+        if err:
+            self.context.nonfatal_error(PostParseError(node.pos, err))
             return None
         else:
             return node
-
-class InterpretCompilerDirectives(CythonTransform):
+    
+class InterpretCompilerDirectives(CythonTransform, SkipDeclarations):
     """
     After parsing, options can be stored in a number of places:
     - #cython-comments at the top of the file (stored in ModuleNode)
@@ -349,10 +409,6 @@ class InterpretCompilerDirectives(CythonTransform):
         else:
             node.cython_attribute = self.option_names.get(node.name)
         return node
-        
-    def visit_Node(self, node):
-        self.visitchildren(node)
-        return node
 
     def try_to_parse_option(self, node):
         # If node is the contents of an option (in a with statement or
@@ -395,7 +451,7 @@ class InterpretCompilerDirectives(CythonTransform):
         return directive
  
     # Handle decorators
-    def visit_DefNode(self, node):
+    def visit_FuncDefNode(self, node):
         options = []
         
         if node.decorators:
@@ -407,7 +463,10 @@ class InterpretCompilerDirectives(CythonTransform):
                     options.append(option)
                 else:
                     realdecs.append(dec)
-            node.decorators = realdecs
+            if realdecs and isinstance(node, CFuncDefNode):
+                raise PostParseError(realdecs[0].pos, "Cdef functions cannot take arbitrary decorators.")
+            else:
+                node.decorators = realdecs
         
         if options:
             optdict = {}
@@ -419,6 +478,16 @@ class InterpretCompilerDirectives(CythonTransform):
             return self.visit_with_options(body, optdict)
         else:
             return self.visit_Node(node)
+    
+    def visit_CVarDefNode(self, node):
+        if node.decorators:
+            for dec in node.decorators:
+                option = self.try_to_parse_option(dec.decorator)
+                if option is not None and option[0] == u'locals':
+                    node.directive_locals = option[1]
+                else:
+                    raise PostParseError(dec.pos, "Cdef functions can only take cython.locals() decorator.")
+        return node
                                    
     # Handle with statements
     def visit_WithStatNode(self, node):
@@ -431,7 +500,7 @@ class InterpretCompilerDirectives(CythonTransform):
         else:
             return self.visit_Node(node)
 
-class WithTransform(CythonTransform):
+class WithTransform(CythonTransform, SkipDeclarations):
 
     # EXCINFO is manually set to a variable that contains
     # the exc_info() tuple that can be generated by the enclosing except
@@ -443,6 +512,7 @@ class WithTransform(CythonTransform):
         EXC = True
         try:
             try:
+                EXCINFO = None
                 BODY
             except:
                 EXC = False
@@ -461,6 +531,7 @@ class WithTransform(CythonTransform):
         EXC = True
         try:
             try:
+                EXCINFO = None
                 TARGET = VALUE
                 BODY
             except:
@@ -470,34 +541,104 @@ class WithTransform(CythonTransform):
         finally:
             if EXC:
                 EXIT(None, None, None)
+            MGR = EXIT = VALUE = EXC = None
+            
     """, temps=[u'MGR', u'EXC', u"EXIT", u"VALUE"],
     pipeline=[NormalizeTree(None)])
 
     def visit_WithStatNode(self, node):
-        excinfo_temp = TempHandle(PyrexTypes.py_object_type)
+        # TODO: Cleanup badly needed
+        TemplateTransform.temp_name_counter += 1
+        handle = "__tmpvar_%d" % TemplateTransform.temp_name_counter
+        
+        self.visitchildren(node, ['body'])
+        excinfo_temp = NameNode(node.pos, name=handle)#TempHandle(Builtin.tuple_type)
         if node.target is not None:
             result = self.template_with_target.substitute({
                 u'EXPR' : node.manager,
                 u'BODY' : node.body,
                 u'TARGET' : node.target,
-                u'EXCINFO' : excinfo_temp.ref(node.pos)
+                u'EXCINFO' : excinfo_temp
                 }, pos=node.pos)
-            # Set except excinfo target to EXCINFO
-            result.body.stats[4].body.stats[0].except_clauses[0].excinfo_target = (
-                excinfo_temp.ref(node.pos))
         else:
             result = self.template_without_target.substitute({
                 u'EXPR' : node.manager,
                 u'BODY' : node.body,
-                u'EXCINFO' : excinfo_temp.ref(node.pos)
+                u'EXCINFO' : excinfo_temp
                 }, pos=node.pos)
-            # Set except excinfo target to EXCINFO
-            result.body.stats[4].body.stats[0].except_clauses[0].excinfo_target = (
-                excinfo_temp.ref(node.pos))
 
-        return TempsBlockNode(node.pos, temps=[excinfo_temp], body=result)
+        # Set except excinfo target to EXCINFO
+        try_except = result.stats[-1].body.stats[-1]
+        try_except.except_clauses[0].excinfo_target = NameNode(node.pos, name=handle)
+#            excinfo_temp.ref(node.pos))
 
-class DecoratorTransform(CythonTransform):
+#        result.stats[-1].body.stats[-1] = TempsBlockNode(
+#            node.pos, temps=[excinfo_temp], body=try_except)
+
+        return result
+        
+    def visit_ExprNode(self, node):
+        # With statements are never inside expressions.
+        return node
+        
+
+class ComprehensionTransform(VisitorTransform):
+    """Prevent the target of list/set/dict comprehensions from leaking by
+    moving it into a temp variable.  This mimics the behaviour of all
+    comprehensions in Py3 and of generator expressions in Py2.x.
+
+    This must run before the IterationTransform, which might replace
+    for-loops with while-loops.  We only handle for-loops here.
+    """
+    def visit_ModuleNode(self, node):
+        self.comprehension_targets = {}
+        self.visitchildren(node)
+        return node
+
+    visit_Node = VisitorTransform.recurse_to_children
+
+    def visit_ComprehensionNode(self, node):
+        if type(node.loop) not in (Nodes.ForInStatNode,
+                                   Nodes.ForFromStatNode):
+            # this should not happen!
+            self.visitchildren(node)
+            return node
+
+        outer_comprehension_targets = self.comprehension_targets
+        self.comprehension_targets = outer_comprehension_targets.copy()
+
+        # find all NameNodes in the loop target
+        target_name_collector = NameNodeCollector()
+        target_name_collector.visit(node.loop.target)
+        targets = target_name_collector.name_nodes
+
+        # create a temp variable for each target name
+        temps = []
+        for target in targets:
+            handle = TempHandle(target.type)
+            temps.append(handle)
+            self.comprehension_targets[target.entry.cname] = handle.ref(node.pos)
+
+        # replace name references in the loop code by their temp node
+        self.visitchildren(node, ['loop'])
+
+        loop = node.loop
+        if type(loop) is Nodes.ForFromStatNode and loop.target.type.is_numeric:
+            loop.loopvar_node = loop.target
+
+        node.loop = TempsBlockNode(node.pos, body=node.loop, temps=temps)
+        self.comprehension_targets = outer_comprehension_targets
+        return node
+
+    def visit_NameNode(self, node):
+        if node.entry:
+            replacement = self.comprehension_targets.get(node.entry.cname)
+            if replacement is not None:
+                return replacement
+        return node
+
+
+class DecoratorTransform(CythonTransform, SkipDeclarations):
 
     def visit_DefNode(self, func_node):
         if not func_node.decorators:
@@ -557,6 +698,21 @@ property NAME:
     # analysis and can be dropped. The analysis was performed
     # on these nodes in a seperate recursive process from the
     # enclosing function or module, so we can simply drop them.
+    def visit_CDeclaratorNode(self, node):
+        return node
+    
+    def visit_CTypeDefNode(self, node):
+        return node
+
+    def visit_CBaseTypeNode(self, node):
+        return None
+    
+    def visit_CEnumDefNode(self, node):
+        return None
+
+    def visit_CStructOrUnionDefNode(self, node):
+        return None
+
     def visit_CVarDefNode(self, node):
         if node.need_properties:
             # cdef public attributes may need type testing on 
@@ -601,6 +757,7 @@ class AlignFunctionDefinitions(CythonTransform):
     
     def visit_ModuleNode(self, node):
         self.scope = node.scope
+        self.directives = node.directives
         self.visitchildren(node)
         return node
     
@@ -613,8 +770,8 @@ class AlignFunctionDefinitions(CythonTransform):
                 error(node.pos, "'%s' redeclared" % node.name)
                 error(pxd_def.pos, "previous declaration here")
                 return None
-        self.visitchildren(node)
-        return node
+        else:
+            return node
         
     def visit_CClassDefNode(self, node, pxd_def=None):
         if pxd_def is None:
@@ -636,6 +793,8 @@ class AlignFunctionDefinitions(CythonTransform):
                 error(node.pos, "'%s' redeclared" % node.name)
                 error(pxd_def.pos, "previous declaration here")
                 return None
+        elif self.scope.is_module_scope and self.directives['auto_cpdef']:
+            node = node.as_cfunction(scope=self.scope)
         # Enable this when internal def functions are allowed. 
         # self.visitchildren(node)
         return node
