@@ -7,7 +7,7 @@ import StringEncoding
 import Naming
 import copy
 
-class BaseType:
+class BaseType(object):
     #
     #  Base class for all Pyrex types including pseudo-types.
 
@@ -20,7 +20,6 @@ class BaseType:
         else:
             return base_code
 
-
 class PyrexType(BaseType):
     #
     #  Base class for all Pyrex types.
@@ -49,7 +48,6 @@ class PyrexType(BaseType):
     #  default_value         string      Initial value
     #  parsetuple_format     string      Format char for PyArg_ParseTuple
     #  pymemberdef_typecode  string      Type code for PyMemberDef struct
-    #  typestring            string      String char defining the type (see Python struct module)
     #
     #  declaration_code(entity_code, 
     #      for_display = 0, dll_linkage = None, pyrex = 0)
@@ -101,7 +99,6 @@ class PyrexType(BaseType):
     default_value = ""
     parsetuple_format = ""
     pymemberdef_typecode = None
-    typestring = None
     
     def resolve(self):
         # If a typedef, returns the base type.
@@ -157,7 +154,6 @@ class CTypedefType(BaseType):
     #  typedef_base_type   PyrexType
     
     is_typedef = 1
-    typestring = None # Because typedefs are not known exactly
     
     def __init__(self, cname, base_type):
         self.typedef_cname = cname
@@ -243,7 +239,6 @@ class PyObjectType(PyrexType):
     parsetuple_format = "O"
     pymemberdef_typecode = "T_OBJECT"
     buffer_defaults = None
-    typestring = "O"
     
     def __str__(self):
         return "Python object"
@@ -261,6 +256,11 @@ class PyObjectType(PyrexType):
         else:
             return "%s *%s" % (public_decl("PyObject", dll_linkage), entity_code)
 
+    def as_pyobject(self, cname):
+        if (not self.is_complete()) or self.is_extension_type:
+            return "(PyObject *)" + cname
+        else:
+            return cname
 
 class BuiltinObjectType(PyObjectType):
 
@@ -303,14 +303,16 @@ class BuiltinObjectType(PyObjectType):
     def type_test_code(self, arg):
         type_name = self.name
         if type_name == 'str':
-            type_name = 'String'
+            check = 'PyString_CheckExact'
         elif type_name == 'set':
-            type_name = 'AnySet'
+            check = 'PyAnySet_CheckExact'
         elif type_name == 'frozenset':
-            type_name = 'FrozenSet'
+            check = 'PyFrozenSet_CheckExact'
+        elif type_name == 'bool':
+            check = 'PyBool_Check'
         else:
-            type_name = type_name.capitalize()
-        return 'likely(Py%s_CheckExact(%s)) || (%s) == Py_None || (PyErr_Format(PyExc_TypeError, "Expected %s, got %%s", Py_TYPE(%s)->tp_name), 0)' % (type_name, arg, arg, self.name, arg)
+            check = 'Py%s_CheckExact' % type_name.capitalize()
+        return 'likely(%s(%s)) || (%s) == Py_None || (PyErr_Format(PyExc_TypeError, "Expected %s, got %%s", Py_TYPE(%s)->tp_name), 0)' % (check, arg, arg, self.name, arg)
 
     def declaration_code(self, entity_code, 
             for_display = 0, dll_linkage = None, pyrex = 0):
@@ -462,17 +464,16 @@ class CNumericType(CType):
     default_value = "0"
     
     parsetuple_formats = ( # rank -> format
-        "BHIkK????", # unsigned
-        "bhilL?fd?", # assumed signed
-        "bhilL?fd?", # explicitly signed
+        "BHIk??K???", # unsigned
+        "bhil??Lfd?", # assumed signed
+        "bhil??Lfd?", # explicitly signed
     )
     
     sign_words = ("unsigned ", "", "signed ")
     
-    def __init__(self, rank, signed = 1, pymemberdef_typecode = None, typestring = None):
+    def __init__(self, rank, signed = 1, pymemberdef_typecode = None):
         self.rank = rank
         self.signed = signed
-        self.typestring = typestring
         ptf = self.parsetuple_formats[signed][rank]
         if ptf == '?':
             ptf = None
@@ -507,9 +508,8 @@ class CIntType(CNumericType):
     from_py_function = "__pyx_PyInt_AsLong"
     exception_value = -1
 
-    def __init__(self, rank, signed, pymemberdef_typecode = None, is_returncode = 0,
-                 typestring=None):
-        CNumericType.__init__(self, rank, signed, pymemberdef_typecode, typestring=typestring)
+    def __init__(self, rank, signed, pymemberdef_typecode = None, is_returncode = 0):
+        CNumericType.__init__(self, rank, signed, pymemberdef_typecode)
         self.is_returncode = is_returncode
         if self.from_py_function == '__pyx_PyInt_AsLong':
             self.from_py_function = self.get_type_conversion()
@@ -519,7 +519,7 @@ class CIntType(CNumericType):
         c_type = self.sign_and_name()
         c_name = c_type.replace(' ', '_');
         func_name = "__pyx_PyInt_%s" % c_name;
-        if not int_conversion_list.has_key(func_name):
+        if func_name not in int_conversion_list:
             # no env to add utility code to
             global type_conversion_predeclarations, type_conversion_functions
             if self.signed:
@@ -560,7 +560,7 @@ class CBIntType(CIntType):
 
 class CAnonEnumType(CIntType):
 
-    is_enum = 1	
+    is_enum = 1    
 
 
 class CUIntType(CIntType):
@@ -595,6 +595,18 @@ class CPySSizeTType(CIntType):
     to_py_function = "PyInt_FromSsize_t"
     from_py_function = "__pyx_PyIndex_AsSsize_t"
 
+    def sign_and_name(self):
+        return rank_to_type_name[self.rank]
+
+
+class CSizeTType(CUIntType):
+
+    to_py_function = "__pyx_PyInt_FromSize_t"
+    from_py_function = "__pyx_PyInt_AsSize_t"
+
+    def sign_and_name(self):
+        return rank_to_type_name[self.rank]
+
 
 class CFloatType(CNumericType):
 
@@ -602,8 +614,8 @@ class CFloatType(CNumericType):
     to_py_function = "PyFloat_FromDouble"
     from_py_function = "__pyx_PyFloat_AsDouble"
     
-    def __init__(self, rank, pymemberdef_typecode = None, typestring=None):
-        CNumericType.__init__(self, rank, 1, pymemberdef_typecode, typestring = typestring)
+    def __init__(self, rank, pymemberdef_typecode = None):
+        CNumericType.__init__(self, rank, 1, pymemberdef_typecode)
     
     def assignable_from_resolved_type(self, src_type):
         return src_type.is_numeric or src_type is error_type
@@ -750,7 +762,7 @@ class CFuncType(CType):
 
     def same_c_signature_as_resolved_type(self, other_type, as_cmethod = 0):
         #print "CFuncType.same_c_signature_as_resolved_type:", \
-        #	self, other_type, "as_cmethod =", as_cmethod ###
+        #    self, other_type, "as_cmethod =", as_cmethod ###
         if other_type is error_type:
             return 1
         if not other_type.is_cfunction:
@@ -782,7 +794,7 @@ class CFuncType(CType):
     
     def compatible_signature_with_resolved_type(self, other_type, as_cmethod):
         #print "CFuncType.same_c_signature_as_resolved_type:", \
-        #	self, other_type, "as_cmethod =", as_cmethod ###
+        #    self, other_type, "as_cmethod =", as_cmethod ###
         if other_type is error_type:
             return 1
         if not other_type.is_cfunction:
@@ -864,7 +876,8 @@ class CFuncType(CType):
             and not (self.nogil and not other_type.nogil)
     
     def declaration_code(self, entity_code, 
-            for_display = 0, dll_linkage = None, pyrex = 0):
+                         for_display = 0, dll_linkage = None, pyrex = 0,
+                         with_calling_convention = 1):
         arg_decl_list = []
         for arg in self.args[:len(self.args)-self.optional_arg_count]:
             arg_decl_list.append(
@@ -890,14 +903,17 @@ class CFuncType(CType):
                 " except *" # ignored
             if self.nogil:
                 trailer += " nogil"
-        cc = self.calling_convention_prefix()
-        if (not entity_code and cc) or entity_code.startswith("*"):
-            entity_code = "(%s%s)" % (cc, entity_code)
-            cc = ""
+        if not with_calling_convention:
+            cc = ''
+        else:
+            cc = self.calling_convention_prefix()
+            if (not entity_code and cc) or entity_code.startswith("*"):
+                entity_code = "(%s%s)" % (cc, entity_code)
+                cc = ""
         return self.return_type.declaration_code(
             "%s%s(%s)%s" % (cc, entity_code, arg_decl_code, trailer),
             for_display, dll_linkage, pyrex)
-	
+    
     def function_header_code(self, func_name, arg_code):
         return "%s%s(%s)" % (self.calling_convention_prefix(),
             func_name, arg_code)
@@ -906,6 +922,10 @@ class CFuncType(CType):
         s = self.declaration_code("")
         return s
 
+    def signature_cast_string(self):
+        s = self.declaration_code("(*)", with_calling_convention=False)
+        return '(%s)' % s
+
 
 class CFuncTypeArg:
     #  name       string
@@ -1130,10 +1150,9 @@ class ErrorType(PyrexType):
     
     is_error = 1
     exception_value = "0"
-    exception_check	= 0
+    exception_check    = 0
     to_py_function = "dummy"
     from_py_function = "dummy"
-    typestring = None
     
     def create_convert_utility_code(self, env):
         return True
@@ -1154,11 +1173,12 @@ rank_to_type_name = (
     "short",        # 1
     "int",          # 2
     "long",         # 3
-    "PY_LONG_LONG", # 4
-    "Py_ssize_t",   # 5
-    "float",        # 6
-    "double",       # 7
-    "long double",  # 8
+    "Py_ssize_t",   # 4
+    "size_t",       # 5
+    "PY_LONG_LONG", # 6
+    "float",        # 7
+    "double",       # 8
+    "long double",  # 9
 )
 
 py_object_type = PyObjectType()
@@ -1167,37 +1187,40 @@ c_void_type =         CVoidType()
 c_void_ptr_type =     CPtrType(c_void_type)
 c_void_ptr_ptr_type = CPtrType(c_void_ptr_type)
 
-c_uchar_type =       CIntType(0, 0, "T_UBYTE", typestring="B")
-c_ushort_type =      CIntType(1, 0, "T_USHORT", typestring="H")
-c_uint_type =        CUIntType(2, 0, "T_UINT", typestring="I")
-c_ulong_type =       CULongType(3, 0, "T_ULONG", typestring="L")
-c_ulonglong_type =   CULongLongType(4, 0, "T_ULONGLONG", typestring="Q")
+c_uchar_type =       CIntType(0, 0, "T_UBYTE")
+c_ushort_type =      CIntType(1, 0, "T_USHORT")
+c_uint_type =        CUIntType(2, 0, "T_UINT")
+c_ulong_type =       CULongType(3, 0, "T_ULONG")
+c_ulonglong_type =   CULongLongType(6, 0, "T_ULONGLONG")
 
-c_char_type =        CIntType(0, 1, "T_CHAR", typestring="b")
-c_short_type =       CIntType(1, 1, "T_SHORT", typestring="h")
-c_int_type =         CIntType(2, 1, "T_INT", typestring="i")
-c_long_type =        CIntType(3, 1, "T_LONG", typestring="l")
-c_longlong_type =    CLongLongType(4, 1, "T_LONGLONG", typestring="q")
-c_py_ssize_t_type =  CPySSizeTType(5, 1)
-c_bint_type =        CBIntType(2, 1, "T_INT", typestring="i")
+c_char_type =        CIntType(0, 1, "T_CHAR")
+c_short_type =       CIntType(1, 1, "T_SHORT")
+c_int_type =         CIntType(2, 1, "T_INT")
+c_long_type =        CIntType(3, 1, "T_LONG")
+c_longlong_type =    CLongLongType(6, 1, "T_LONGLONG")
+c_bint_type =        CBIntType(2, 1, "T_INT")
 
-c_schar_type =       CIntType(0, 2, "T_CHAR", typestring="b")
-c_sshort_type =      CIntType(1, 2, "T_SHORT", typestring="h")
-c_sint_type =        CIntType(2, 2, "T_INT", typestring="i")
-c_slong_type =       CIntType(3, 2, "T_LONG", typestring="l")
-c_slonglong_type =   CLongLongType(4, 2, "T_LONGLONG", typestring="q")
+c_schar_type =       CIntType(0, 2, "T_CHAR")
+c_sshort_type =      CIntType(1, 2, "T_SHORT")
+c_sint_type =        CIntType(2, 2, "T_INT")
+c_slong_type =       CIntType(3, 2, "T_LONG")
+c_slonglong_type =   CLongLongType(6, 2, "T_LONGLONG")
 
-c_float_type =       CFloatType(6, "T_FLOAT", typestring="f")
-c_double_type =      CFloatType(7, "T_DOUBLE", typestring="d")
-c_longdouble_type =  CFloatType(8, typestring="g")
+c_py_ssize_t_type =  CPySSizeTType(4, 2, "T_PYSSIZET")
+c_size_t_type =      CSizeTType(5, 0, "T_SIZET")
+
+c_float_type =       CFloatType(7, "T_FLOAT")
+c_double_type =      CFloatType(8, "T_DOUBLE")
+c_longdouble_type =  CFloatType(9)
 
 c_null_ptr_type =     CNullPtrType(c_void_type)
 c_char_array_type =   CCharArrayType(None)
 c_char_ptr_type =     CCharPtrType()
 c_utf8_char_array_type = CUTF8CharArrayType(None)
 c_char_ptr_ptr_type = CPtrType(c_char_ptr_type)
-c_py_ssize_t_ptr_type =  CPtrType(c_py_ssize_t_type)
 c_int_ptr_type =      CPtrType(c_int_type)
+c_py_ssize_t_ptr_type =  CPtrType(c_py_ssize_t_type)
+c_size_t_ptr_type =  CPtrType(c_size_t_type)
 
 c_returncode_type =   CIntType(2, 1, "T_INT", is_returncode = 1)
 
@@ -1210,35 +1233,40 @@ c_py_buffer_ptr_type = CPtrType(c_py_buffer_type)
 error_type =    ErrorType()
 unspecified_type = UnspecifiedType()
 
-lowest_float_rank = 6
-
 sign_and_rank_to_type = {
     #(signed, rank)
-    (0, 0, ): c_uchar_type, 
+    (0, 0): c_uchar_type, 
     (0, 1): c_ushort_type, 
     (0, 2): c_uint_type, 
-  (0, 3): c_ulong_type,
-  (0, 4): c_ulonglong_type,
-    (0, 5):  c_ulonglong_type,            # I'm not sure about this.  this should be for size_t Py_ssize_t
+    (0, 3): c_ulong_type,
+    (0, 6): c_ulonglong_type,
+
     (1, 0): c_char_type, 
     (1, 1): c_short_type, 
     (1, 2): c_int_type, 
     (1, 3): c_long_type,
-    (1, 4): c_longlong_type,
-    (1, 5): c_py_ssize_t_type,
+    (1, 6): c_longlong_type,
+
     (2, 0): c_schar_type, 
     (2, 1): c_sshort_type, 
     (2, 2): c_sint_type, 
     (2, 3): c_slong_type,
-    (2, 4): c_slonglong_type,
-    (2, 5): c_py_ssize_t_type,
-    (1, 6): c_float_type, 
-    (1, 7): c_double_type,
-    (1, 8): c_longdouble_type,
+    (2, 6): c_slonglong_type, 
+
+    (0, 4): c_py_ssize_t_type,
+    (1, 4): c_py_ssize_t_type,
+    (2, 4): c_py_ssize_t_type,
+    (0, 5): c_size_t_type,
+    (1, 5): c_size_t_type,
+    (2, 5): c_size_t_type,
+
+    (1, 7): c_float_type, 
+    (1, 8): c_double_type,
+    (1, 9): c_longdouble_type,
 # In case we're mixing unsigned ints and floats...
-    (0, 6): c_float_type, 
-    (0, 7): c_double_type,
-    (0, 8): c_longdouble_type,
+    (0, 7): c_float_type, 
+    (0, 8): c_double_type,
+    (0, 9): c_longdouble_type,
 }
 
 modifiers_and_name_to_type = {
@@ -1254,7 +1282,6 @@ modifiers_and_name_to_type = {
     (1, 0, "int"): c_int_type, 
     (1, 1, "int"): c_long_type,
     (1, 2, "int"): c_longlong_type,
-    (1, 0, "Py_ssize_t"): c_py_ssize_t_type,
     (1, 0, "float"): c_float_type, 
     (1, 0, "double"): c_double_type,
     (1, 1, "double"): c_longdouble_type,
@@ -1265,8 +1292,10 @@ modifiers_and_name_to_type = {
     (2, 0, "int"): c_sint_type, 
     (2, 1, "int"): c_slong_type,
     (2, 2, "int"): c_slonglong_type,
+
     (2, 0, "Py_ssize_t"): c_py_ssize_t_type,
-    
+    (0, 0, "size_t") : c_size_t_type,
+
     (1, 0, "long"): c_long_type,
     (1, 0, "short"): c_short_type,
     (1, 0, "longlong"): c_longlong_type,
@@ -1368,18 +1397,51 @@ type_conversion_predeclarations = """
 /* Type Conversion Predeclarations */
 
 #if PY_MAJOR_VERSION < 3
-#define __Pyx_PyBytes_FromString PyString_FromString
-#define __Pyx_PyBytes_AsString   PyString_AsString
+#define __Pyx_PyBytes_FromString          PyString_FromString
+#define __Pyx_PyBytes_FromStringAndSize   PyString_FromStringAndSize
+#define __Pyx_PyBytes_AsString            PyString_AsString
 #else
-#define __Pyx_PyBytes_FromString PyBytes_FromString
-#define __Pyx_PyBytes_AsString   PyBytes_AsString
+#define __Pyx_PyBytes_FromString          PyBytes_FromString
+#define __Pyx_PyBytes_FromStringAndSize   PyBytes_FromStringAndSize
+#define __Pyx_PyBytes_AsString            PyBytes_AsString
 #endif
 
 #define __Pyx_PyBool_FromLong(b) ((b) ? (Py_INCREF(Py_True), Py_True) : (Py_INCREF(Py_False), Py_False))
 static INLINE int __Pyx_PyObject_IsTrue(PyObject* x);
 static INLINE PY_LONG_LONG __pyx_PyInt_AsLongLong(PyObject* x);
 static INLINE unsigned PY_LONG_LONG __pyx_PyInt_AsUnsignedLongLong(PyObject* x);
+
+#if !defined(T_PYSSIZET)
+#if PY_VERSION_HEX < 0x02050000
+#define T_PYSSIZET T_INT
+#elif !defined(T_LONGLONG)
+#define T_PYSSIZET \\
+        ((sizeof(Py_ssize_t) == sizeof(int))  ? T_INT  : \\
+        ((sizeof(Py_ssize_t) == sizeof(long)) ? T_LONG : -1))
+#else
+#define T_PYSSIZET \\
+        ((sizeof(Py_ssize_t) == sizeof(int))          ? T_INT      : \\
+        ((sizeof(Py_ssize_t) == sizeof(long))         ? T_LONG     : \\
+        ((sizeof(Py_ssize_t) == sizeof(PY_LONG_LONG)) ? T_LONGLONG : -1)))
+#endif
+#endif
+
+#if !defined(T_SIZET)
+#if !defined(T_ULONGLONG)
+#define T_SIZET \\
+        ((sizeof(size_t) == sizeof(unsigned int))  ? T_UINT  : \\
+        ((sizeof(size_t) == sizeof(unsigned long)) ? T_ULONG : -1))
+#else
+#define T_SIZET \\
+        ((sizeof(size_t) == sizeof(unsigned int))          ? T_UINT      : \\
+        ((sizeof(size_t) == sizeof(unsigned long))         ? T_ULONG     : \\
+        ((sizeof(size_t) == sizeof(unsigned PY_LONG_LONG)) ? T_ULONGLONG : -1)))
+#endif
+#endif
+
 static INLINE Py_ssize_t __pyx_PyIndex_AsSsize_t(PyObject* b);
+static INLINE PyObject * __pyx_PyInt_FromSize_t(size_t);
+static INLINE size_t __pyx_PyInt_AsSize_t(PyObject*);
 
 #define __pyx_PyInt_AsLong(x) (PyInt_CheckExact(x) ? PyInt_AS_LONG(x) : PyInt_AsLong(x))
 #define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
@@ -1397,44 +1459,85 @@ static INLINE Py_ssize_t __pyx_PyIndex_AsSsize_t(PyObject* b) {
   return ival;
 }
 
+static INLINE PyObject * __pyx_PyInt_FromSize_t(size_t ival) {
+#if PY_VERSION_HEX < 0x02050000
+   if (ival <= LONG_MAX)
+       return PyInt_FromLong((long)ival);
+   else {
+       unsigned char *bytes = (unsigned char *) &ival;
+       int one = 1; int little = (int)*(unsigned char*)&one;
+       return _PyLong_FromByteArray(bytes, sizeof(size_t), little, 0);
+   }
+#else
+   return PyInt_FromSize_t(ival);
+#endif
+}
+
+static INLINE size_t __pyx_PyInt_AsSize_t(PyObject* b) {
+   unsigned PY_LONG_LONG val = __pyx_PyInt_AsUnsignedLongLong(b);
+   if (unlikely(val == (unsigned PY_LONG_LONG)-1 && PyErr_Occurred())) {
+       return (size_t)-1;
+   } else if (unlikely(val != (unsigned PY_LONG_LONG)(size_t)val)) {
+       PyErr_SetString(PyExc_OverflowError, "value too large to convert to size_t");
+       return (size_t)-1;
+   }
+   return val;
+}
+
 static INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
    if (x == Py_True) return 1;
-   else if (x == Py_False) return 0;
+   else if ((x == Py_False) | (x == Py_None)) return 0;
    else return PyObject_IsTrue(x);
 }
 
 static INLINE PY_LONG_LONG __pyx_PyInt_AsLongLong(PyObject* x) {
+#if PY_VERSION_HEX < 0x03000000
     if (PyInt_CheckExact(x)) {
         return PyInt_AS_LONG(x);
     }
-    else if (PyLong_CheckExact(x)) {
+    else
+#endif
+    if (PyLong_CheckExact(x)) {
         return PyLong_AsLongLong(x);
     }
     else {
         PY_LONG_LONG val;
-        PyObject* tmp = PyNumber_Int(x); if (!tmp) return (PY_LONG_LONG)-1;
+#if PY_VERSION_HEX < 0x03000000
+        PyObject* tmp = PyNumber_Int(x);  if (!tmp) return (PY_LONG_LONG)-1;
         val = __pyx_PyInt_AsLongLong(tmp);
+#else
+        PyObject* tmp = PyNumber_Long(x); if (!tmp) return (PY_LONG_LONG)-1;
+        val = PyLong_AsLongLong(tmp);
+#endif
         Py_DECREF(tmp);
         return val;
     }
 }
 
 static INLINE unsigned PY_LONG_LONG __pyx_PyInt_AsUnsignedLongLong(PyObject* x) {
+#if PY_VERSION_HEX < 0x03000000
     if (PyInt_CheckExact(x)) {
         long val = PyInt_AS_LONG(x);
         if (unlikely(val < 0)) {
-            PyErr_SetString(PyExc_TypeError, "Negative assignment to unsigned type.");
+            PyErr_SetString(PyExc_OverflowError, "can't convert negative value to unsigned long long");
             return (unsigned PY_LONG_LONG)-1;
         }
         return val;
     }
-    else if (PyLong_CheckExact(x)) {
+    else
+#endif
+    if (PyLong_CheckExact(x)) {
         return PyLong_AsUnsignedLongLong(x);
     }
     else {
-        PY_LONG_LONG val;
-        PyObject* tmp = PyNumber_Int(x); if (!tmp) return (PY_LONG_LONG)-1;
+        unsigned PY_LONG_LONG val;
+#if PY_VERSION_HEX < 0x03000000
+        PyObject* tmp = PyNumber_Int(x);  if (!tmp) return (PY_LONG_LONG)-1;
         val = __pyx_PyInt_AsUnsignedLongLong(tmp);
+#else
+        PyObject* tmp = PyNumber_Long(x); if (!tmp) return (PY_LONG_LONG)-1;
+        val = PyLong_AsUnsignedLongLong(tmp);
+#endif
         Py_DECREF(tmp);
         return val;
     }
