@@ -68,6 +68,7 @@ class Entry:
     # in_closure       boolean    Is referenced in an inner scope
     # is_readonly      boolean    Can't be assigned to
     # func_cname       string     C func implementing Python func
+    # func_modifiers   [string]   C function modifiers ('inline')
     # pos              position   Source position where declared
     # namespace_cname  string     If is_pyglobal, the C variable
     #                               holding its home namespace
@@ -96,7 +97,10 @@ class Entry:
     # utility_code     string     Utility code needed when this entry is used
     #
     # buffer_aux      BufferAux or None  Extra information needed for buffer variables
+    # inline_func_in_pxd boolean  Hacky special case for inline function in pxd file.
+    #                             Ideally this should not be necesarry.
 
+    inline_func_in_pxd = False
     borrowed = 0
     init = ""
     visibility = 'private'
@@ -122,6 +126,7 @@ class Entry:
     is_declared_generic = 0
     is_readonly = 0
     func_cname = None
+    func_modifiers = []
     doc = None
     init_to_none = 0
     as_variable = None
@@ -277,8 +282,8 @@ class Scope:
         if not self.in_cinclude and cname and re.match("^_[_A-Z]+$", cname):
             # See http://www.gnu.org/software/libc/manual/html_node/Reserved-Names.html#Reserved-Names 
             warning(pos, "'%s' is a reserved name in C." % cname, -1)
-        dict = self.entries
-        if name and dict.has_key(name):
+        entries = self.entries
+        if name and name in entries:
             if visibility == 'extern':
                 warning(pos, "'%s' redeclared " % name, 0)
             elif visibility != 'ignore':
@@ -287,7 +292,7 @@ class Scope:
         entry.in_cinclude = self.in_cinclude
         if name:
             entry.qualified_name = self.qualify_name(name)
-            dict[name] = entry
+            entries[name] = entry
         entry.scope = self
         entry.visibility = visibility
         return entry
@@ -298,7 +303,7 @@ class Scope:
     def declare_const(self, name, type, value, pos, cname = None, visibility = 'private'):
         # Add an entry for a named constant.
         if not cname:
-            if self.in_cinclude:
+            if self.in_cinclude or visibility == 'public':
                 cname = name
             else:
                 cname = self.mangle(Naming.enum_prefix, name)
@@ -387,7 +392,7 @@ class Scope:
             visibility = visibility)
         entry.enum_values = []
         self.sue_entries.append(entry)
-        return entry	
+        return entry    
     
     def declare_var(self, name, type, pos, 
             cname = None, visibility = 'private', is_cdef = 0):
@@ -420,7 +425,8 @@ class Scope:
         self.pyfunc_entries.append(entry)
     
     def declare_cfunction(self, name, type, pos, 
-            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
+                          cname = None, visibility = 'private', defining = 0,
+                          api = 0, in_pxd = 0, modifiers = ()):
         # Add an entry for a C function.
         entry = self.lookup_here(name)
         if entry:
@@ -438,7 +444,7 @@ class Scope:
                     cname = name
                 else:
                     cname = self.mangle(Naming.func_prefix, name)
-            entry = self.add_cfunction(name, type, pos, cname, visibility)
+            entry = self.add_cfunction(name, type, pos, cname, visibility, modifiers)
             entry.func_cname = cname
         if in_pxd and visibility != 'extern':
             entry.defined_in_pxd = 1
@@ -448,12 +454,16 @@ class Scope:
             error(pos, "Non-extern C function '%s' declared but not defined" % name)
         if defining:
             entry.is_implemented = True
+        if modifiers:
+            entry.func_modifiers = modifiers
         return entry
     
-    def add_cfunction(self, name, type, pos, cname, visibility):
+    def add_cfunction(self, name, type, pos, cname, visibility, modifiers):
         # Add a C function entry without giving it a func_cname.
         entry = self.declare(name, cname, type, pos, visibility)
         entry.is_cfunction = 1
+        if modifiers:
+            entry.func_modifiers = modifiers
         self.cfunc_entries.append(entry)
         return entry
     
@@ -652,7 +662,7 @@ class Scope:
         # Test whether any of the given names are
         # defined in this scope.
         for name in names:
-            if name in self.entries:	
+            if name in self.entries:    
                 return 1
         return 0
 
@@ -795,7 +805,12 @@ class ModuleScope(Scope):
         self.parent_module = parent_module
         outer_scope = context.find_submodule("__builtin__")
         Scope.__init__(self, name, outer_scope, parent_module)
-        self.module_name = name
+        if name != "__init__":
+            self.module_name = name
+        else:
+            # Treat Spam/__init__.pyx specially, so that when Python loads
+            # Spam/__init__.so, initSpam() is defined.
+            self.module_name = parent_module.module_name
         self.context = context
         self.module_cname = Naming.module_cname
         self.module_dict_cname = Naming.moddict_cname
@@ -821,6 +836,8 @@ class ModuleScope(Scope):
         self.cached_builtins = []
         self.undeclared_cached_builtins = []
         self.namespace_cname = self.module_cname
+        for name in ['__builtins__', '__name__', '__file__', '__doc__']:
+            self.declare_var(EncodedString(name), py_object_type, None)
     
     def qualifying_scope(self):
         return self.parent_module
@@ -887,7 +904,11 @@ class ModuleScope(Scope):
     
     def add_imported_module(self, scope):
         if scope not in self.cimported_modules:
+            for filename in scope.include_files:
+                self.add_include_file(filename)
             self.cimported_modules.append(scope)
+            for m in scope.cimported_modules:
+                self.add_imported_module(m)
     
     def add_imported_entry(self, name, entry, pos):
         if entry not in self.entries:
@@ -918,7 +939,7 @@ class ModuleScope(Scope):
         else:
             entry = self.declare_var(name, py_object_type, pos)
         entry.as_module = scope
-        self.cimported_modules.append(scope)
+        self.add_imported_module(scope)
         return entry
     
     def declare_var(self, name, type, pos, 
@@ -1007,7 +1028,7 @@ class ModuleScope(Scope):
             if objstruct_cname:
                 type.objstruct_cname = objstruct_cname
             elif not entry.in_cinclude:
-                type.objstruct_cname = self.mangle(Naming.objstruct_prefix, name)				
+                type.objstruct_cname = self.mangle(Naming.objstruct_prefix, name)                
             else:
                 error(entry.pos, 
                     "Object name required for 'public' or 'extern' C class")
@@ -1046,13 +1067,13 @@ class ModuleScope(Scope):
         if objstruct_cname:
             if type.objstruct_cname and type.objstruct_cname != objstruct_cname:
                 error(pos, "Object struct name differs from previous declaration")
-            type.objstruct_cname = objstruct_cname		
+            type.objstruct_cname = objstruct_cname        
         if typeobj_cname:
             if type.typeobj_cname and type.typeobj_cname != typeobj_cname:
                     error(pos, "Type object name differs from previous declaration")
             type.typeobj_cname = typeobj_cname
         #
-        # Return new or existing entry	
+        # Return new or existing entry    
         #
         return entry
     
@@ -1245,7 +1266,8 @@ class StructOrUnionScope(Scope):
         return entry
 
     def declare_cfunction(self, name, type, pos, 
-            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
+                          cname = None, visibility = 'private', defining = 0,
+                          api = 0, in_pxd = 0, modifiers = ()):
         self.declare_var(name, type, pos, cname, visibility)
 
 class ClassScope(Scope):
@@ -1307,7 +1329,7 @@ class PyClassScope(ClassScope):
         self.outer_scope.release_temp(cname)
 
     #def recycle_pending_temps(self):
-    #	self.outer_scope.recycle_pending_temps()
+    #    self.outer_scope.recycle_pending_temps()
 
     def add_default_value(self, type):
         return self.outer_scope.add_default_value(type)
@@ -1426,7 +1448,8 @@ class CClassScope(ClassScope):
         return ClassScope.lookup_here(self, name)
     
     def declare_cfunction(self, name, type, pos,
-            cname = None, visibility = 'private', defining = 0, api = 0, in_pxd = 0):
+                          cname = None, visibility = 'private',
+                          defining = 0, api = 0, in_pxd = 0, modifiers = ()):
         if get_special_method_signature(name):
             error(pos, "Special methods must be declared with 'def', not 'cdef'")
         args = type.args
@@ -1445,7 +1468,7 @@ class CClassScope(ClassScope):
                 if type.same_c_signature_as(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
                     pass
                 elif type.compatible_signature_with(entry.type, as_cmethod = 1) and type.nogil == entry.type.nogil:
-                    entry = self.add_cfunction(name, type, pos, cname or name, visibility='ignore')
+                    entry = self.add_cfunction(name, type, pos, cname or name, visibility='ignore', modifiers=modifiers)
                     defining = 1
                 else:
                     error(pos, "Signature not compatible with previous declaration")
@@ -1455,15 +1478,17 @@ class CClassScope(ClassScope):
                 error(pos,
                     "C method '%s' not previously declared in definition part of"
                     " extension type" % name)
-            entry = self.add_cfunction(name, type, pos, cname or name, visibility)
+            entry = self.add_cfunction(name, type, pos, cname or name,
+                                       visibility, modifiers)
         if defining:
             entry.func_cname = self.mangle(Naming.func_prefix, name)
         return entry
         
-    def add_cfunction(self, name, type, pos, cname, visibility):
+    def add_cfunction(self, name, type, pos, cname, visibility, modifiers):
         # Add a cfunction entry without giving it a func_cname.
         prev_entry = self.lookup_here(name)
-        entry = ClassScope.add_cfunction(self, name, type, pos, cname, visibility)
+        entry = ClassScope.add_cfunction(self, name, type, pos, cname,
+                                         visibility, modifiers)
         entry.is_cmethod = 1
         entry.prev_entry = prev_entry
         return entry
@@ -1494,7 +1519,8 @@ class CClassScope(ClassScope):
                 self.inherited_var_entries.append(entry)
         for base_entry in base_scope.cfunc_entries:
             entry = self.add_cfunction(base_entry.name, base_entry.type,
-                    base_entry.pos, adapt(base_entry.cname), base_entry.visibility)
+                                       base_entry.pos, adapt(base_entry.cname),
+                                       base_entry.visibility, base_entry.func_modifiers)
             entry.is_inherited = 1
             
     def allocate_temp(self, type):
@@ -1536,7 +1562,7 @@ impl = """
 static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
     /* It appears that PyMethodDescr_Type is not anywhere exposed in the Python/C API */
     /* if (!PyObject_TypeCheck(method, &PyMethodDescr_Type)) { */ 
-    if (strcmp(Py_TYPE(method)->tp_name, "method_descriptor") == 0) { /* cdef classes */
+    if (__Pyx_StrEq(Py_TYPE(method)->tp_name, "method_descriptor")) { /* cdef classes */
         PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
         return PyDescr_NewClassMethod(descr->d_type, descr->d_method);
     }
