@@ -1,10 +1,12 @@
 #
 #   Tree visitor and transform framework
 #
+import cython
 import inspect
 import Nodes
 import ExprNodes
 import Naming
+import Errors
 from StringEncoding import EncodedString
 
 class BasicVisitor(object):
@@ -14,26 +16,33 @@ class BasicVisitor(object):
     # instance.
     def __init__(self):
         self.dispatch_table = {}
-     
+
     def visit(self, obj):
-        cls = obj.__class__
-        m = self.dispatch_table.get(cls.__name__)
-        if m is None:
+        cls = type(obj)
+        try:
+            handler_method = self.dispatch_table[cls]
+        except KeyError:
+            #print "Cache miss for class %s in visitor %s" % (
+            #    cls.__name__, type(self).__name__)
             # Must resolve, try entire hierarchy
             pattern = "visit_%s"
             mro = inspect.getmro(cls)
-            for cls in mro:
-                m = getattr(self, pattern % cls.__name__, None)
-                if m is not None:
+            handler_method = None
+            for mro_cls in mro:
+                if hasattr(self, pattern % mro_cls.__name__):
+                    handler_method = getattr(self, pattern % mro_cls.__name__)
                     break
-            else:
+            if handler_method is None:
                 print type(self), type(obj)
-                print self.access_path
-                print self.access_path[-1][0].pos
-                print self.access_path[-1][0].__dict__
+                if hasattr(self, 'access_path') and self.access_path:
+                    print self.access_path
+                    if self.access_path:
+                        print self.access_path[-1][0].pos
+                        print self.access_path[-1][0].__dict__
                 raise RuntimeError("Visitor does not accept object: %s" % obj)
-            self.dispatch_table[cls.__name__] = m
-        return m(obj)
+            #print "Caching " + cls.__name__
+            self.dispatch_table[cls] = handler_method
+        return handler_method(obj)
 
 class TreeVisitor(BasicVisitor):
     """
@@ -53,7 +62,7 @@ class TreeVisitor(BasicVisitor):
     
     Example:
     
-    >>> class SampleNode:
+    >>> class SampleNode(object):
     ...     child_attrs = ["head", "body"]
     ...     def __init__(self, value, head=None, body=None):
     ...         self.value = value
@@ -83,9 +92,85 @@ class TreeVisitor(BasicVisitor):
         super(TreeVisitor, self).__init__()
         self.access_path = []
 
+    def dump_node(self, node, indent=0):
+        ignored = list(node.child_attrs) + [u'child_attrs', u'pos',
+                                            u'gil_message', u'subexprs']
+        values = []
+        pos = node.pos
+        if pos:
+            source = pos[0]
+            if source:
+                import os.path
+                source = os.path.basename(source.get_description())
+            values.append(u'%s:%s:%s' % (source, pos[1], pos[2]))
+        attribute_names = dir(node)
+        attribute_names.sort()
+        for attr in attribute_names:
+            if attr in ignored:
+                continue
+            if attr.startswith(u'_') or attr.endswith(u'_'):
+                continue
+            try:
+                value = getattr(node, attr)
+            except AttributeError:
+                continue
+            if value is None or value == 0:
+                continue
+            elif isinstance(value, list):
+                value = u'[...]/%d' % len(value)
+            elif not isinstance(value, (str, unicode, long, int, float)):
+                continue
+            else:
+                value = repr(value)
+            values.append(u'%s = %s' % (attr, value))
+        return u'%s(%s)' % (node.__class__.__name__,
+                           u',\n    '.join(values))
+
+    def _find_node_path(self, stacktrace):
+        import os.path
+        last_traceback = stacktrace
+        nodes = []
+        while hasattr(stacktrace, 'tb_frame'):
+            frame = stacktrace.tb_frame
+            node = frame.f_locals.get(u'self')
+            if isinstance(node, Nodes.Node):
+                code = frame.f_code
+                method_name = code.co_name
+                pos = (os.path.basename(code.co_filename),
+                       code.co_firstlineno)
+                nodes.append((node, method_name, pos))
+                last_traceback = stacktrace
+            stacktrace = stacktrace.tb_next
+        return (last_traceback, nodes)
+
     def visitchild(self, child, parent, attrname, idx):
         self.access_path.append((parent, attrname, idx))
-        result = self.visit(child)
+        try:
+            result = self.visit(child)
+        except Errors.CompileError:
+            raise
+        except Exception, e:
+            import sys
+            trace = ['']
+            for parent, attribute, index in self.access_path:
+                node = getattr(parent, attribute)
+                if index is None:
+                    index = ''
+                else:
+                    node = node[index]
+                    index = u'[%d]' % index
+                trace.append(u'%s.%s%s = %s' % (
+                    parent.__class__.__name__, attribute, index,
+                    self.dump_node(node)))
+            stacktrace, called_nodes = self._find_node_path(sys.exc_info()[2])
+            last_node = child
+            for node, method_name, pos in called_nodes:
+                last_node = node
+                trace.append(u"File '%s', line %d, in %s: %s" % (
+                    pos[0], pos[1], method_name, self.dump_node(node)))
+            raise Errors.CompilerCrash(
+                last_node.pos, self.__class__.__name__,
+                u'\n'.join(trace), e, stacktrace)
         self.access_path.pop()
         return result
 
@@ -135,8 +220,13 @@ class VisitorTransform(TreeVisitor):
     was not, an exception will be raised. (Typically you want to ensure that you
     are within a StatListNode or similar before doing this.)
     """
+    def __init__(self):
+        super(VisitorTransform, self).__init__()
+        self._super_visitchildren = super(VisitorTransform, self).visitchildren
+
     def visitchildren(self, parent, attrs=None):
-        result = super(VisitorTransform, self).visitchildren(parent, attrs)
+        result = cython.declare(dict)
+        result = self._super_visitchildren(parent, attrs)
         for attr, newnode in result.iteritems():
             if not isinstance(newnode, list):
                 setattr(parent, attr, newnode)
@@ -151,6 +241,10 @@ class VisitorTransform(TreeVisitor):
                             newlist.append(x)
                 setattr(parent, attr, newlist)
         return result        
+
+    def recurse_to_children(self, node):
+        self.visitchildren(node)
+        return node
     
     def __call__(self, root):
         return self.visit(root)
