@@ -22,6 +22,7 @@ try:
                compile as cython_compile
     from Cython.Compiler.Errors import PyrexError
 except ImportError, e:
+    print "failed to import Cython: %s" % e
     PyrexError = None
 
 from distutils.command import build_ext as _build_ext
@@ -85,9 +86,9 @@ class build_ext(_build_ext.build_ext):
 
         """
         Walk the list of source files in 'sources', looking for Cython
-        source (.pyx) files.  Run Cython on all that are found, and return
-        a modified 'sources' list with Cython source files replaced by the
-        generated C (or C++) files.
+        source files (.pyx and .py).  Run Cython on all that are
+        found, and return a modified 'sources' list with Cython source
+        files replaced by the generated C (or C++) files.
         """
 
         if PyrexError == None:
@@ -105,17 +106,17 @@ class build_ext(_build_ext.build_ext):
         # cplus will also be set to true is extension.language is equal to
         # 'C++' or 'c++'.
         #try:
-        #	create_listing = self.pyrex_create_listing or \
-        #						extension.pyrex_create_listing
-        #	cplus = self.pyrex_cplus or \
-        #				extension.pyrex_cplus or \
-        #				(extension.language != None and \
-        #					extension.language.lower() == 'c++')
+        #    create_listing = self.pyrex_create_listing or \
+        #                        extension.pyrex_create_listing
+        #    cplus = self.pyrex_cplus or \
+        #                extension.pyrex_cplus or \
+        #                (extension.language != None and \
+        #                    extension.language.lower() == 'c++')
         #except AttributeError:
-        #	create_listing = self.pyrex_create_listing
-        #	cplus = self.pyrex_cplus or \
-        #				(extension.language != None and \
-        #					extension.language.lower() == 'c++')
+        #    create_listing = self.pyrex_create_listing
+        #    cplus = self.pyrex_cplus or \
+        #                (extension.language != None and \
+        #                    extension.language.lower() == 'c++')
         
         create_listing = self.pyrex_create_listing or \
             getattr(extension, 'pyrex_create_listing', 0)
@@ -124,10 +125,10 @@ class build_ext(_build_ext.build_ext):
         pyrex_gen_pxi = self.pyrex_gen_pxi or getattr(extension, 'pyrex_gen_pxi', 0)
 
         # Set up the include_path for the Cython compiler:
-        #	1.	Start with the command line option.
-        #	2.	Add in any (unique) paths from the extension
-        #		pyrex_include_dirs (if Cython.Distutils.extension is used).
-        #	3.	Add in any (unique) paths from the extension include_dirs
+        #    1.    Start with the command line option.
+        #    2.    Add in any (unique) paths from the extension
+        #        pyrex_include_dirs (if Cython.Distutils.extension is used).
+        #    3.    Add in any (unique) paths from the extension include_dirs
         includes = self.pyrex_include_dirs
         try:
             for i in extension.pyrex_include_dirs:
@@ -158,7 +159,10 @@ class build_ext(_build_ext.build_ext):
         newest_dependency = None
         for source in sources:
             (base, ext) = os.path.splitext(os.path.basename(source))
-            if ext == ".pyx":			  # Cython source file
+            if ext == ".py":
+                # FIXME: we might want to special case this some more
+                ext = '.pyx'
+            if ext == ".pyx":              # Cython source file
                 output_dir = target_dir or os.path.dirname(source)
                 new_sources.append(os.path.join(output_dir, base + target_ext))
                 pyrex_sources.append(source)
